from .auth import (
    Authenticator,
    KeyfileAuthenticator,
    PrivateKeyAuthenticator,
    TrezorAuthenticator,
)
from .config import Config
from .api.admin import Admin
from .api.margin_account import MarginAccount
from .api.product import Product
from .api.trading import Trading
from .constants import defaults
from .exceptions import ConfigurationError
from .validators import validate_address


class AFP:
    """Application object for interacting with the AFP Clearing System and the AutEx
    exchange.

    Configuration precedence:
    1) Defaults: Autonity Mainnet parameters, or Autonity Bakerloo (Nile) Testnet
       parameters when `AFP_TESTNET=true` is set.
    2) Environment variables override defaults.
    3) AFP constructor arguments override environment variables.
    4) Admin, MarginAccount, Product, and Trading API constructor arguments override
       AFP constructor arguments.

    Parameters
    ----------
    authenticator : afp.Authenticator, optional
        The default authenticator for signing transactions & messages. Can also be set
        with environment variables: use `AFP_PRIVATE_KEY` for private key
        authentication; `AFP_KEYFILE` and `AFP_KEYFILE_PASSWORD` for keyfile
        authentication; `AFP_TREZOR_PATH_OR_INDEX` and `AFP_TREZOR_PASSPHRASE` for
        Trezor device authentication.
    rpc_url : str, optional
        The URL of an Autonity RPC provider. Can also be set with the `AFP_RPC_URL`
        environment variable.
    exchange_url : str, optional
        The REST API base URL of the exchange. Defaults to the URL of the AutEx
        exchange. Its default value can be overridden with the `AFP_EXCHANGE_URL`
        environment variable.
    ipfs_api_url : str, optional
        The RPC API root URL of an IPFS node that supports Kubo RPC API v0, required
        for product registration. Defaults to the URL of a local IPFS node. Its default
        value can be overridden with the `AFP_IPFS_API_URL` environment variable.
    ipfs_api_key : str, optional
        The access token to include in API requests to the IPFS client. Its default
        value can be overridden with the `AFP_IPFS_API_KEY` environment variable.
    chain_id : str, optional
        The chain ID of the Autonity network. Defauls to the chain ID of Autonity
        Mainnet. Its default value can be overridden with the `AFP_CHAIN_ID` environment
        variable.
    gas_limit : int, optional
        The `gasLimit` parameter of blockchain transactions. Estimated with the
        `eth_estimateGas` JSON-RPC call if not specified. Its default value can be
        overridden with the `AFP_GAS_LIMIT` environment variable.
    max_fee_per_gas : int, optional
        The `maxFeePerGas` parameter of blockchain transactions in ton (wei) units.
        Defaults to `baseFeePerGas` from the return value of the `eth_getBlock` JSON-RPC
        call. Its default value can be overridden with the `AFP_MAX_FEE_PER_GAS`
        environment variable.
    max_priority_fee_per_gas : int, optional
        The `maxPriorityFeePerGas` parameter of blockchain transactions in ton (wei)
        units. Defaults to the return value of the `eth_maxPriorityFeePerGas` JSON-RPC
        call. Its default value can be overridden with the
        `AFP_MAX_PRIORITY_FEE_PER_GAS` environment variable.
    timeout_seconds: int, optional
        The number of seconds to wait for a blockchain transaction to be mined.
        Defaults to 10 seconds. Its default value can be overridden with the
        `AFP_TIMEOUT_SECONDS` environment variable.
    clearing_diamond_address : str, optional
        The address of the ClearingDiamond contract. Defaults to the Autonity Mainnet
        deployment address. Its default value can be overridden with the
        `AFP_CLEARING_DIAMOND_ADDRESS` environment variable.
    margin_account_registry_address : str, optional
        The address of the MarginAccountRegistry contract. Defaults to the Autonity
        Mainnet deployment address. Its default value can be overridden with the
        `AFP_MARGIN_ACCOUNT_REGISTRY_ADDRESS` environment variable.
    oracle_provider_address : str, optional
        The address of the OracleProvider contract. Defaults to the Autonity Mainnet
        deployment address. Its default value can be overridden with the
        `AFP_ORACLE_PROVIDER_ADDRESS` environment variable.
    product_registry_address: str, optional
        The address of the ProductRegistry contract. Defaults to the Autonity Mainnet
        deployment address. Its default value can be overridden with the
        `AFP_PRODUCT_REGISTRY_ADDRESS` environment variable.
    system_viewer_address: str, optional
        The address of the SystemViewer contract. Defaults to the Autonity Mainnet
        deployment address. Its default value can be overridden with the
        `AFP_SYSTEM_VIEWER_ADDRESS` environment variable.
    """

    config: Config

    def __init__(
        self,
        *,
        authenticator: Authenticator | None = None,
        rpc_url: str | None = defaults.RPC_URL,
        exchange_url: str = defaults.EXCHANGE_URL,
        ipfs_api_url: str = defaults.IPFS_API_URL,
        ipfs_api_key: str | None = defaults.IPFS_API_KEY,
        chain_id: int = defaults.CHAIN_ID,
        gas_limit: int | None = defaults.GAS_LIMIT,
        max_fee_per_gas: int | None = defaults.MAX_FEE_PER_GAS,
        max_priority_fee_per_gas: int | None = defaults.MAX_PRIORITY_FEE_PER_GAS,
        timeout_seconds: int = defaults.TIMEOUT_SECONDS,
        clearing_diamond_address: str = defaults.CLEARING_DIAMOND_ADDRESS,
        margin_account_registry_address: str = defaults.MARGIN_ACCOUNT_REGISTRY_ADDRESS,
        oracle_provider_address: str = defaults.ORACLE_PROVIDER_ADDRESS,
        product_registry_address: str = defaults.PRODUCT_REGISTRY_ADDRESS,
        system_viewer_address: str = defaults.SYSTEM_VIEWER_ADDRESS,
    ) -> None:
        if authenticator is None:
            authenticator = _default_authenticator()

        self.config = Config(
            authenticator=authenticator,
            exchange_url=exchange_url,
            rpc_url=rpc_url,
            ipfs_api_url=ipfs_api_url,
            ipfs_api_key=ipfs_api_key,
            chain_id=chain_id,
            gas_limit=gas_limit,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas,
            timeout_seconds=timeout_seconds,
            clearing_diamond_address=validate_address(clearing_diamond_address),
            margin_account_registry_address=validate_address(
                margin_account_registry_address
            ),
            oracle_provider_address=validate_address(oracle_provider_address),
            product_registry_address=validate_address(product_registry_address),
            system_viewer_address=validate_address(system_viewer_address),
        )

    def __repr__(self) -> str:
        return f"AFP(config={repr(self.config)})"

    # Clearing APIs

    def MarginAccount(
        self, authenticator: Authenticator | None = None
    ) -> MarginAccount:
        """API for managing margin accounts.

        Parameters
        ----------
        authenticator : afp.Authenticator, optional
            Authenticator for signing transactions sent to the Clearing System.
            Defaults to the authenticator specified in the `AFP` constructor.
        """
        return MarginAccount(self.config, authenticator=authenticator)

    def Product(self, authenticator: Authenticator | None = None) -> Product:
        """API for managing products.

        Parameters
        ----------
        authenticator : afp.Authenticator, optional
            Authenticator for signing transactions sent to the Clearing System.
            Defaults to the authenticator specified in the `AFP` constructor.
        """
        return Product(self.config, authenticator=authenticator)

    # Exchange APIs

    def Admin(
        self,
        authenticator: Authenticator | None = None,
        exchange_url: str | None = None,
    ) -> Admin:
        """API for AutEx administration, restricted to AutEx admins.

        Authenticates with the exchange on creation.

        Parameters
        ----------
        authenticator : afp.Authenticator, optional
            Authenticator for authenticating with the AutEx exchange. Defaults to the
            authenticator specified in the `AFP` constructor.
        exchange_url: str, optional
            The REST API base URL of the exchange. Defaults to the value specified in
            the `AFP` constructor.

        Raises
        ------
        afp.exceptions.AuthenticationError
            If the exchange rejects the login attempt.
        """
        return Admin(
            self.config, authenticator=authenticator, exchange_url=exchange_url
        )

    def Trading(
        self,
        authenticator: Authenticator | None = None,
        exchange_url: str | None = None,
    ) -> Trading:
        """API for trading in the AutEx exchange.

        Authenticates with the exchange on creation.

        Parameters
        ----------
        authenticator : afp.Authenticator, optional
            Authenticator for signing intents and authenticating with the AutEx
            exchange. Defaults to the authenticator specified in the `AFP` constructor.
        exchange_url: str, optional
            The REST API base URL of the exchange. Defaults to the value specified in
            the `AFP` constructor.

        Raises
        ------
        afp.exceptions.AuthenticationError
            If the exchange rejects the login attempt.
        """
        return Trading(
            self.config, authenticator=authenticator, exchange_url=exchange_url
        )


def _default_authenticator() -> Authenticator | None:
    auth_variable_count = sum(
        [
            int(bool(defaults.PRIVATE_KEY)),
            int(bool(defaults.KEYFILE)),
            int(bool(defaults.TREZOR_PATH_OR_INDEX)),
        ]
    )
    if auth_variable_count > 1:
        raise ConfigurationError(
            "Only one of AFP_PRIVATE_KEY, AFP_KEYFILE and AFP_TREZOR_PATH_OR_INDEX "
            "environment variables should be specified"
        )

    if defaults.PRIVATE_KEY:
        return PrivateKeyAuthenticator(defaults.PRIVATE_KEY)
    if defaults.KEYFILE:
        return KeyfileAuthenticator(defaults.KEYFILE, defaults.KEYFILE_PASSWORD)
    if defaults.TREZOR_PATH_OR_INDEX:
        return TrezorAuthenticator(
            defaults.TREZOR_PATH_OR_INDEX, defaults.TREZOR_PASSPHRASE
        )
    return None



from itertools import chain

from eth_typing.evm import ChecksumAddress

from web3 import Web3

from . import (
    admin_facet,
    clearing_facet,
    final_settlement_facet,
    margin_account_facet,
    margin_account_registry,
    mark_price_tracker_facet,
    oracle_provider,
    product_registry,
    product_registry_facet,
    system_viewer,
)
from ..constants import defaults

# In order to include a facet in the ClearingDiamond facade:
# 1. Add its ABI to CLEARING_DIAMOND_ABI
# 2. Set its contract binding as a superclass of ClearingDiamond

CLEARING_DIAMOND_ABI = list(
    chain(
        admin_facet.ABI,
        clearing_facet.ABI,
        final_settlement_facet.ABI,
        margin_account_facet.ABI,
        mark_price_tracker_facet.ABI,
        product_registry_facet.ABI,
    )
)


class ClearingDiamond(
    admin_facet.AdminFacet,
    clearing_facet.ClearingFacet,
    final_settlement_facet.FinalSettlementFacet,
    margin_account_facet.MarginAccountFacet,
    mark_price_tracker_facet.MarkPriceTrackerFacet,
    product_registry_facet.ProductRegistryFacet,
):
    """ClearingDiamond contract binding.

    Includes all functions inherited from various facets.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(
        self, w3: Web3, address: ChecksumAddress = defaults.CLEARING_DIAMOND_ADDRESS
    ):
        self._contract = w3.eth.contract(address=address, abi=CLEARING_DIAMOND_ABI)


class MarginAccountRegistry(margin_account_registry.MarginAccountRegistry):
    """MarginAccountRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(
        self,
        w3: Web3,
        address: ChecksumAddress = defaults.MARGIN_ACCOUNT_REGISTRY_ADDRESS,
    ):
        super().__init__(w3, address)


class OracleProvider(oracle_provider.OracleProvider):
    """OracleProvider contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(
        self, w3: Web3, address: ChecksumAddress = defaults.ORACLE_PROVIDER_ADDRESS
    ):
        super().__init__(w3, address)


class ProductRegistry(product_registry.ProductRegistry):
    """ProductRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(
        self, w3: Web3, address: ChecksumAddress = defaults.PRODUCT_REGISTRY_ADDRESS
    ):
        super().__init__(w3, address)


class SystemViewer(system_viewer.SystemViewer):
    """SystemViewer contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    """

    def __init__(
        self, w3: Web3, address: ChecksumAddress = defaults.SYSTEM_VIEWER_ADDRESS
    ):
        super().__init__(w3, address)



import atexit
import json
import os
import sys
from typing import Protocol, cast

import trezorlib.ethereum as trezor_eth
from eth_account.account import Account
from eth_account.datastructures import SignedTransaction
from eth_account.messages import encode_defunct
from eth_account.signers.local import LocalAccount
from eth_account.types import TransactionDictType
from eth_account._utils.legacy_transactions import (
    encode_transaction,
    serializable_unsigned_transaction_from_dict,
)
from eth_typing.evm import ChecksumAddress
from eth_utils.conversions import to_int
from eth_utils.crypto import keccak
from hexbytes import HexBytes
from trezorlib.client import TrezorClient
from trezorlib.client import get_default_client  # type: ignore
from trezorlib.ui import TrezorClientUI
from trezorlib.tools import parse_path
from trezorlib.transport import DeviceIsBusy
from web3 import Web3
from web3.constants import CHECKSUM_ADDRESSS_ZERO
from web3.types import TxParams

from .constants import TREZOR_DEFAULT_PREFIX
from .exceptions import DeviceError


class Authenticator(Protocol):
    address: ChecksumAddress

    def sign_message(self, message: bytes) -> HexBytes: ...

    def sign_transaction(self, params: TxParams) -> SignedTransaction: ...


class NullAuthenticator(Authenticator):
    """Authenticator stub as placeholder for testing."""

    def __init__(self):
        self.address = CHECKSUM_ADDRESSS_ZERO

    def sign_message(self, message: bytes) -> HexBytes:
        raise NotImplementedError()

    def sign_transaction(self, params: TxParams) -> SignedTransaction:
        raise NotImplementedError()


class PrivateKeyAuthenticator(Authenticator):
    """Authenticates with a private key specified in a constructor argument.

    Parameters
    ----------
    private_key: str
        The private key of a blockchain account.
    """

    _account: LocalAccount

    def __init__(self, private_key: str) -> None:
        self._account = Account.from_key(private_key)
        self.address = self._account.address

    def sign_message(self, message: bytes) -> HexBytes:
        eip191_message = encode_defunct(message)
        signed_message = self._account.sign_message(eip191_message)
        return signed_message.signature

    def sign_transaction(self, params: TxParams) -> SignedTransaction:
        return self._account.sign_transaction(cast(TransactionDictType, params))

    def __repr__(self):
        return f"{self.__class__.__name__}(address='{self.address}')"


class KeyfileAuthenticator(PrivateKeyAuthenticator):
    """Authenticates with a private key read from an encrypted keyfile.

    Parameters
    ----------
    key_file : str
        The path to the keyfile.
    password : str
        The password for decrypting the keyfile. Defaults to no password.
    """

    def __init__(self, key_file: str, password: str = "") -> None:
        with open(os.path.expanduser(key_file), encoding="utf8") as f:
            key_data = json.load(f)

        private_key = Account.decrypt(key_data, password=password)
        super().__init__(private_key.to_0x_hex())


class TrezorAuthenticator(Authenticator):
    """Authenticates with a Trezor device.

    Parameters
    ----------
    path_or_index: str or int
        The full derivation path of the account, e.g. `m/44'/60'/0'/0/123`; or the
        index of the account at the default Trezor derivation prefix for Ethereum
        accounts `m/44'/60'/0'/0`, e.g. `123`.
    passphrase: str
        The passphrase for the Trezor device. Defaults to no passphrase.
    """

    client: TrezorClient[TrezorClientUI]

    def __init__(self, path_or_index: str | int, passphrase: str = ""):
        if isinstance(path_or_index, int) or path_or_index.isdigit():
            path_str = f"{TREZOR_DEFAULT_PREFIX}/{int(path_or_index)}"
        else:
            path_str = path_or_index.replace("'", "h")
        try:
            self.path = parse_path(path_str)
        except ValueError as exc:
            raise DeviceError(
                f"Invalid Trezor BIP32 derivation path '{path_str}'"
            ) from exc
        self.client = self._get_client(passphrase)
        atexit.register(self.client.end_session)

        address_str = trezor_eth.get_address(  # type: ignore
            self.client, self.path
        )
        self.address = Web3.to_checksum_address(address_str)

    def sign_transaction(self, params: TxParams) -> SignedTransaction:
        assert "chainId" in params
        assert "gas" in params
        assert "nonce" in params
        assert "to" in params
        assert "value" in params
        data_bytes = HexBytes(params["data"] if "data" in params else b"")

        print("[Confirm on Trezor device]", file=sys.stderr)

        if "gasPrice" in params and params["gasPrice"]:
            v_int, r_bytes, s_bytes = trezor_eth.sign_tx(  # type: ignore
                self.client,
                self.path,
                nonce=cast(int, params["nonce"]),
                gas_price=cast(int, params["gasPrice"]),
                gas_limit=params["gas"],
                to=cast(str, params["to"]),
                value=cast(int, params["value"]),
                data=data_bytes,
                chain_id=params["chainId"],
            )
        else:
            assert "maxFeePerGas" in params
            assert "maxPriorityFeePerGas" in params
            v_int, r_bytes, s_bytes = trezor_eth.sign_tx_eip1559(  # type: ignore
                self.client,
                self.path,
                nonce=cast(int, params["nonce"]),
                gas_limit=params["gas"],
                to=cast(str, params["to"]),
                value=cast(int, params["value"]),
                data=data_bytes,
                chain_id=params["chainId"],
                max_gas_fee=int(params["maxFeePerGas"]),
                max_priority_fee=int(params["maxPriorityFeePerGas"]),
            )

        r_int = to_int(r_bytes)
        s_int = to_int(s_bytes)
        filtered_tx = dict((k, v) for (k, v) in params.items() if k not in ("from"))
        # In a LegacyTransaction, "type" is not a valid field. See EIP-2718.
        if "type" in filtered_tx and filtered_tx["type"] == "0x0":
            filtered_tx.pop("type")
        tx_unsigned = serializable_unsigned_transaction_from_dict(
            cast(TransactionDictType, filtered_tx)
        )
        tx_encoded = encode_transaction(tx_unsigned, vrs=(v_int, r_int, s_int))
        txhash = keccak(tx_encoded)
        return SignedTransaction(
            raw_transaction=HexBytes(tx_encoded),
            hash=HexBytes(txhash),
            r=r_int,
            s=s_int,
            v=v_int,
        )

    def sign_message(self, message: bytes) -> HexBytes:
        print("[Confirm on Trezor device]", file=sys.stderr)

        sigdata = trezor_eth.sign_message(  # type: ignore
            self.client,
            self.path,
            message.decode("utf-8"),
        )
        return HexBytes(sigdata.signature)

    @staticmethod
    def _get_client(
        passphrase: str,
    ) -> TrezorClient[TrezorClientUI]:
        ui = _NonInteractiveTrezorUI(passphrase)
        try:
            return cast(
                TrezorClient[TrezorClientUI],
                get_default_client(ui=ui),
            )
        except DeviceIsBusy as exc:
            raise DeviceError("Device in use by another process") from exc
        except Exception as exc:
            raise DeviceError(
                "No Trezor device found; "
                "check device is connected, unlocked, and detected by OS"
            ) from exc


class _NonInteractiveTrezorUI(TrezorClientUI):
    """Replacement for the default TrezorClientUI of the Trezor library.

    Bringing up an interactive passphrase prompt is unwanted in the SDK;
    this implementation receives the passphrase as constructor argument.
    """

    _passphrase: str

    def __init__(self, passphrase: str) -> None:
        self._passphrase = passphrase

    def button_request(self, br: object) -> None:
        pass

    def get_pin(self, code: object) -> str:
        raise DeviceError("PIN entry on host is not supported")

    def get_passphrase(self, available_on_device: bool) -> str:
        return self._passphrase



import os
from datetime import timedelta
from importlib import metadata
from types import SimpleNamespace


def _int_or_none(value: str | None) -> int | None:
    return int(value) if value is not None else None


# Venue API constants
USER_AGENT = "afp-sdk/{}".format(metadata.version("afp-sdk"))
DEFAULT_BATCH_SIZE = 50
DEFAULT_EXCHANGE_API_VERSION = 1

# Clearing System constants
RATE_MULTIPLIER = 10**4
FEE_RATE_MULTIPLIER = 10**6
FULL_PRECISION_MULTIPLIER = 10**18

# IPFS client constants
IPFS_CID_ENCODING = "base32"
IPFS_REQUEST_TIMEOUT = 30
IPFS_LOCAL_NODE_URL = "http://localhost:5001"
JSON_SCHEMAS_DIRECTORY = os.path.join(os.path.dirname(__file__), "json-schemas")

# Product schema constants
MINIMUM_ORACLE_FALLBACK_INTERVAL = timedelta(days=7)
MONTH_CODES = {
    "F": 1,
    "G": 2,
    "H": 3,
    "J": 4,
    "K": 5,
    "M": 6,
    "N": 7,
    "Q": 8,
    "U": 9,
    "V": 10,
    "X": 11,
    "Z": 12,
}

# Trezor device constants
TREZOR_DEFAULT_PREFIX = "m/44h/60h/0h/0"

schema_cids = SimpleNamespace(
    # afp-product-schemas v0.2.0
    ORACLE_CONFIG_V020="bafyreifcec2km7hxwq6oqzjlspni2mgipetjb7pqtaewh2efislzoctboi",
    ORACLE_CONFIG_PROTOTYPE1_V020="bafyreiaw34o6l3rmatabzbds2i2myazdw2yolevcpsoyd2i2g3ms7wa2eq",
    ORACLE_FALLBACK_V020="bafyreicgr6dfo5yduixjkcifghiulskfegwojvuwodtouvivl362zndhxe",
    OUTCOME_POINT_V020="bafyreibnfg6nq74dvpkre5rakkccij7iadp5rxpim7omsatjnrpmj3y7v4",
    OUTCOME_POINT_EVENT_V020="bafyreihur3dzwhja6uxsbcw6eeoj3xmmc4e3zkmyzpot5v5dleevxe5zam",
    OUTCOME_POINT_TIME_SERIES_V020=(
        "bafyreidzs7okcpqiss6ztftltyptqwnw5e5opsy5yntospekjha4kpykaa"
    ),
    OUTCOME_SPACE_V020="bafyreicheoypx6synljushh7mq2572iyhlolf4nake2p5dwobgnj3r5eua",
    OUTCOME_SPACE_SCALAR_V020="bafyreihn3oiaxffe4e2w7pwtreadpw3obfd7gqlogbcxm56jc2hzfvco74",
    OUTCOME_SPACE_TIME_SERIES_V020=(
        "bafyreid35a67db4sqh4fs6boddyt2xvscbqy6nqvsp5jjur56qhkw4ixre"
    ),
)

mainnet = SimpleNamespace(
    EXCHANGE_URL="https://afp-exchange-stable.up.railway.app/",
    CHAIN_ID=65000000,
    CLEARING_DIAMOND_ADDRESS="0x08e725d38BCABCf62806E84bE044dF67E8b476Ae",
    MARGIN_ACCOUNT_REGISTRY_ADDRESS="0x1FAa342Ea34332ee1186FE48f03894F70F7fFc4f",
    ORACLE_PROVIDER_ADDRESS="0x06CaDDDf6CC08048596aE051c8ce644725219C73",
    PRODUCT_REGISTRY_ADDRESS="0xB605fc6B76eA4592CaA3AEd26EFd3edbAc5A3A24",
    SYSTEM_VIEWER_ADDRESS="0xBeC22149650CcaC2a1233c7b489aFF3ea24df36d",
)

bakerloo = SimpleNamespace(
    EXCHANGE_URL="https://exchange-server-staging.up.railway.app/",
    CHAIN_ID=65010004,
    CLEARING_DIAMOND_ADDRESS="0x6Dc8cd65B03e0462C4D5E954b924ec86D3408FE7",
    MARGIN_ACCOUNT_REGISTRY_ADDRESS="0x801cCa84bcb57418044c882Dfb4B03cc35bE848f",
    ORACLE_PROVIDER_ADDRESS="0x72EeD9f7286292f119089F56e3068a3A931FCD49",
    PRODUCT_REGISTRY_ADDRESS="0xab792A6e84c5A22C7A27FEd904AB83ca2fbA4c5e",
    SYSTEM_VIEWER_ADDRESS="0xF2F903B8956Ca6868E165989A9ebEEE72F4D3e3F",
)

TESTNET = os.getenv("AFP_TESTNET") in ("1", "true", "True")
_current_env = bakerloo if TESTNET else mainnet

defaults = SimpleNamespace(
    # Authentication parameters
    KEYFILE=os.getenv("AFP_KEYFILE", None),
    KEYFILE_PASSWORD=os.getenv("AFP_KEYFILE_PASSWORD", ""),
    PRIVATE_KEY=os.getenv("AFP_PRIVATE_KEY", None),
    TREZOR_PATH_OR_INDEX=os.getenv("AFP_TERZOR_PATH_OR_INDEX", None),
    TREZOR_PASSPHRASE=os.getenv("AFP_TREZOR_PASSPHRASE", ""),
    # Venue parameters
    EXCHANGE_URL=os.getenv("AFP_EXCHANGE_URL", _current_env.EXCHANGE_URL),
    # IPFS client parameters
    IPFS_API_URL=os.getenv("AFP_IPFS_API_URL", IPFS_LOCAL_NODE_URL),
    IPFS_API_KEY=os.getenv("AFP_IPFS_API_KEY", None),
    # Blockchain parameters
    RPC_URL=os.getenv("AFP_RPC_URL", None),
    CHAIN_ID=int(os.getenv("AFP_CHAIN_ID", _current_env.CHAIN_ID)),
    GAS_LIMIT=_int_or_none(os.getenv("AFP_GAS_LIMIT", None)),
    MAX_FEE_PER_GAS=_int_or_none(os.getenv("AFP_MAX_FEE_PER_GAS", None)),
    MAX_PRIORITY_FEE_PER_GAS=_int_or_none(
        os.getenv("AFP_MAX_PRIORITY_FEE_PER_GAS", None)
    ),
    TIMEOUT_SECONDS=int(os.getenv("AFP_TIMEOUT_SECONDS", 10)),
    # Clearing System parameters
    CLEARING_DIAMOND_ADDRESS=os.getenv(
        "AFP_CLEARING_DIAMOND_ADDRESS", _current_env.CLEARING_DIAMOND_ADDRESS
    ),
    MARGIN_ACCOUNT_REGISTRY_ADDRESS=os.getenv(
        "AFP_MARGIN_ACCOUNT_REGISTRY_ADDRESS",
        _current_env.MARGIN_ACCOUNT_REGISTRY_ADDRESS,
    ),
    ORACLE_PROVIDER_ADDRESS=os.getenv(
        "AFP_ORACLE_PROVIDER_ADDRESS",
        _current_env.ORACLE_PROVIDER_ADDRESS,
    ),
    PRODUCT_REGISTRY_ADDRESS=os.getenv(
        "AFP_PRODUCT_REGISTRY_ADDRESS", _current_env.PRODUCT_REGISTRY_ADDRESS
    ),
    SYSTEM_VIEWER_ADDRESS=os.getenv(
        "AFP_SYSTEM_VIEWER_ADDRESS", _current_env.SYSTEM_VIEWER_ADDRESS
    ),
)



from .. import validators
from ..decorators import refresh_token_on_expiry
from ..enums import ListingState
from ..dtos import ExchangeProductListingSubmission, ExchangeProductUpdateSubmission
from .base import ExchangeAPI


class Admin(ExchangeAPI):
    """API for AutEx administration, restricted to AutEx admins."""

    @refresh_token_on_expiry
    def list_product(self, product_id: str) -> None:
        """Lists a product on the exchange.

        The product is in private state after listing until it is revealed to the public.

        Parameters
        ----------
        product_id : str

        Raises
        ------
        afp.exceptions.AuthorizationError
            If the configured account is not an exchange administrator.
        """
        product_listing = ExchangeProductListingSubmission(id=product_id)
        self._exchange.list_product(product_listing)

    @refresh_token_on_expiry
    def reveal_product(self, product_id: str) -> None:
        """Makes a product publicly available for trading on the exchange.

        Parameters
        ----------
        product_id : str

        Raises
        ------
        afp.exceptions.AuthorizationError
            If the configured account is not an exchange administrator.
        """
        product_update = ExchangeProductUpdateSubmission(
            listing_state=ListingState.PUBLIC
        )
        self._exchange.update_product_listing(
            validators.validate_hexstr32(product_id), product_update
        )

    @refresh_token_on_expiry
    def delist_product(self, product_id: str) -> None:
        """Delists a product from the exchange.

        New order submissions of this product will be rejected.

        Parameters
        ----------
        product_id : str

        Raises
        ------
        afp.exceptions.AuthorizationError
            If the configured account is not an exchange administrator.
        """
        product_update = ExchangeProductUpdateSubmission(
            listing_state=ListingState.DELISTED
        )
        self._exchange.update_product_listing(
            validators.validate_hexstr32(product_id), product_update
        )



from datetime import date, datetime, UTC
from functools import partial
from typing import Annotated, Any, ClassVar, Self

import inflection
import multiformats
import rfc8785
from pydantic import (
    AfterValidator,
    AliasGenerator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PlainSerializer,
    model_validator,
)

from . import constants, validators

CID_MODEL_MAP: dict[str, type["PinnedModel"]] = {}


# Conversions


def ensure_py_datetime(value: datetime | int | float | str) -> datetime:
    if isinstance(value, datetime):
        return value.astimezone(UTC)
    if isinstance(value, int) or isinstance(value, float):
        return datetime.fromtimestamp(value, UTC)
    return datetime.fromisoformat(value).astimezone(UTC)


def ensure_py_date(value: date | str) -> date:
    if isinstance(value, date):
        return value
    return date.fromisoformat(value)


def ensure_timestamp(value: datetime) -> int:
    return int(value.timestamp())


def ensure_iso_datetime(value: datetime) -> str:
    assert value.tzinfo is UTC, f"{value} should be in UTC timezone"
    return value.strftime("%Y-%m-%dT%H:%M:%SZ")


def ensure_iso_date(value: date) -> str:
    return value.isoformat()


# Custom types


URL = Annotated[
    str,
    Field(min_length=1, max_length=2083),
    BeforeValidator(validators.validate_url),
    AfterValidator(validators.verify_url),
]

CID = Annotated[
    str,
    Field(
        pattern=r"^(Qm[1-9A-HJ-NP-Za-km-z]{44}|b[a-z2-7]{58,}|z[1-9A-HJ-NP-Za-km-z]{48,})$"
    ),
]

# Decode CIDs for serialization so that the DAG-CBOR encoder will convert them
# into IPLD Link format
IPLD_LINK = Annotated[
    CID,
    BeforeValidator(  # Ensure the same encoding is used consistently
        lambda cid: multiformats.CID.decode(str(cid)).encode(
            constants.IPFS_CID_ENCODING
        )
    ),
    PlainSerializer(multiformats.CID.decode, return_type=multiformats.CID),
]

# Use datetime internally but UNIX timestamp in serialized format
Timestamp = Annotated[
    datetime,
    BeforeValidator(ensure_py_datetime),
    AfterValidator(validators.validate_non_negative_timestamp),
    PlainSerializer(ensure_timestamp, return_type=int),
]

# Use datetime internally but ISO string in serialized format
ISODateTime = Annotated[
    datetime,
    BeforeValidator(ensure_py_datetime),
    AfterValidator(validators.validate_non_negative_timestamp),
    PlainSerializer(ensure_iso_datetime, return_type=str),
]

# Use date internally but ISO string in serialized format
ISODate = Annotated[
    date,
    BeforeValidator(ensure_py_date),
    PlainSerializer(ensure_iso_date, return_type=str),
]


# Base models


class Model(BaseModel):
    """Base immutable schema."""

    model_config = ConfigDict(frozen=True)

    # Always serialize/deserialize by alias

    def model_dump(self, by_alias: bool = True, **kwargs: Any) -> dict[Any, Any]:
        return super().model_dump(by_alias=by_alias, **kwargs)

    def model_dump_json(self, by_alias: bool = True, **kwargs: Any) -> str:
        return super().model_dump_json(by_alias=by_alias, **kwargs)

    def model_dump_canonical_json(self, **kwargs: Any) -> str:
        obj = self.model_dump(mode="json", **kwargs)
        return rfc8785.dumps(obj).decode("utf-8")

    @classmethod
    def model_validate(cls, *args: Any, by_alias: bool = True, **kwargs: Any) -> Self:
        return super().model_validate(*args, by_alias=by_alias, **kwargs)

    @classmethod
    def model_validate_json(
        cls, *args: Any, by_alias: bool = True, **kwargs: Any
    ) -> Self:
        return super().model_validate_json(*args, by_alias=by_alias, **kwargs)


class AliasedModel(Model):
    """Schema that converts property names from snake case to camel case for
    serialization.
    """

    model_config = Model.model_config | ConfigDict(
        alias_generator=AliasGenerator(
            alias=partial(inflection.camelize, uppercase_first_letter=False),
        ),
        populate_by_name=True,
    )


class PinnedModel(Model):
    """Extended metadata schema that has an IPFS CID."""

    SCHEMA_CID: ClassVar[CID]

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        if "SCHEMA_CID" in cls.__dict__:
            assert cls.SCHEMA_CID not in CID_MODEL_MAP, (
                f"{cls.__name__} model does not have unique CID"
            )
        CID_MODEL_MAP[cls.SCHEMA_CID] = cls

    @model_validator(mode="after")
    def _ensure_schema_cid(self) -> Self:
        assert "SCHEMA_CID" in self.__class__.__dict__, (
            f"SCHEMA_CID is missing from {self.__class__.__name__} schema"
        )
        return self



import re
from datetime import date, datetime, timedelta
from decimal import Decimal
from functools import reduce
from operator import getitem
from typing import Any, Iterable

import requests
from binascii import Error
from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from pydantic import AnyUrl
from web3 import Web3

from .constants import MINIMUM_ORACLE_FALLBACK_INTERVAL, MONTH_CODES
from .exceptions import ValidationError


# Generic type validators


def validate_timedelta(value: timedelta) -> timedelta:
    if value.total_seconds() < 0:
        raise ValueError(f"{value} should be positive")
    return value


def validate_non_negative_timestamp(value: datetime) -> datetime:
    if value.timestamp() < 0:
        raise ValueError(f"{value} should be greater than {datetime.fromtimestamp(0)}")
    return value


def validate_hexstr(value: str, length: int | None = None) -> str:
    value = str(value)
    if not value.startswith("0x"):
        raise ValueError(f"{value} should start with '0x'")
    try:
        byte_value = HexBytes(value)
    except Error:
        raise ValueError(f"{value} includes non-hexadecimal characters")
    if length is not None and len(byte_value) != length:
        raise ValueError(f"{value} should be 32 bytes long")
    return value


def validate_hexstr32(value: str) -> str:
    return validate_hexstr(value, length=32)


def validate_address(value: str) -> ChecksumAddress:
    try:
        return Web3.to_checksum_address(value)
    except ValueError:
        raise ValueError(f"{value} is not a valid blockchain address")


def validate_url(value: str) -> str:
    AnyUrl(value)
    return value


def validate_all_caps(value: str) -> str:
    if value != value.upper():
        raise ValueError(f"{value} should use capital letters only")
    return value


# Live data verifiers


def verify_collateral_asset(w3: Web3, address: str) -> ChecksumAddress:
    address = validate_address(address)
    if len(w3.eth.get_code(address)) == 0:
        raise ValidationError(
            f"No contract found at collateral asset address {address}"
        )
    return address


def verify_oracle(w3: Web3, address: str) -> ChecksumAddress:
    address = validate_address(address)
    if len(w3.eth.get_code(address)) == 0:
        raise ValidationError(f"No contract found at oracle address {address}")
    return address


def verify_url(value: str) -> str:
    try:
        requests.head(value)
    except requests.RequestException as ex:
        raise ValidationError(f"Not possible to connect to URL {value}") from ex
    return value


# Schema-specific model validators


def validate_limit_price(
    value: Decimal,
    min_price: Decimal,
    max_price: Decimal,
    tick_size: int,
    rounding: str | None = None,
) -> Decimal:
    if value < min_price:
        raise ValueError(
            f"IntentData: limit_price: {value} should be greater than "
            f"the product's minimum price {min_price}"
        )
    if value > max_price:
        raise ValueError(
            f"IntentData: limit_price: {value} should be less than the product's "
            f"maximum price {max_price}"
        )
    if rounding is None:
        num_fractional_digits = abs(int(value.normalize().as_tuple().exponent))
        if num_fractional_digits > tick_size:
            raise ValueError(
                f"IntentData: limit_price: {value} should have at most "
                f"{tick_size} fractional digits"
            )
    return value.quantize(Decimal("10") ** -tick_size, rounding=rounding)


def validate_price_limits(min_price: Decimal, max_price: Decimal) -> None:
    if min_price >= max_price:
        raise ValueError(
            f"PredictionProductV1.min_price {min_price} "
            f"should be less than PredictionProductV1.max_price {max_price}"
        )


def validate_time_limits(
    start_time: datetime, earliest_fsp_submission_time: datetime
) -> None:
    if start_time >= earliest_fsp_submission_time:
        raise ValueError(
            f"BaseProduct.start_time {start_time} should be less than "
            "ExpirySpecification.earliest_fsp_submission_time "
            f"{earliest_fsp_submission_time}"
        )


def validate_matching_fsp_types(
    outcome_space_fsp_type: str, outcome_point_fsp_type: str
) -> None:
    if outcome_space_fsp_type != outcome_point_fsp_type:
        raise ValueError(
            f"OutcomeSpace.fsp_type '{outcome_space_fsp_type}' should match "
            f"OutcomePoint.fsp_type '{outcome_point_fsp_type}'"
        )


def validate_oracle_fallback_time(
    fallback_time: datetime, earliest_fsp_submission_time: datetime
) -> None:
    earliest_fallback_time = (
        earliest_fsp_submission_time + MINIMUM_ORACLE_FALLBACK_INTERVAL
    )
    if fallback_time < earliest_fallback_time:
        raise ValueError(
            f"OracleFallback: fallback_time {fallback_time} "
            f"should be at least {MINIMUM_ORACLE_FALLBACK_INTERVAL} after "
            "ExpirySpecification.earliest_fsp_submission_time "
            f"{earliest_fsp_submission_time}"
        )


def validate_oracle_fallback_fsp(
    fallback_fsp: Decimal, min_price: Decimal, max_price: Decimal
) -> None:
    if not (min_price <= fallback_fsp <= max_price):
        raise ValueError(
            f"OracleFallback: fallback_fsp {fallback_fsp} should be between "
            f"PredictionProductV1.min_price {min_price} and "
            f"PredictionProductV1.max_price {max_price}"
        )


def validate_outcome_space_template_variables(
    values: Iterable[str], outcome_point_dict: dict[Any, Any]
) -> None:
    for value in values:
        for variable in re.findall(r"{(.*?)}", value):
            try:
                referred_value = reduce(
                    getitem, variable.split("."), outcome_point_dict
                )
            except (TypeError, KeyError):
                raise ValueError(
                    f"OutcomeSpace: Invalid template variable '{variable}'"
                )
            if isinstance(referred_value, dict) or isinstance(referred_value, list):  # type: ignore
                raise ValueError(
                    f"OutcomeSpace: Template variable '{variable}' "
                    "should not refer to a nested object or list"
                )


def validate_symbol(symbol: str, frequency: str, release_date: date) -> None:
    match frequency:
        case "daily":
            pattern = (
                r"[A-Z]{1,8}(?P<day>(0[1-9]|[1-2][0-9]|3[01]))"
                r"(?P<month>[FGHJKMNQUVXZ])(?P<year>[0-9]{2})"
            )
        case "weekly" | "fortnightly":
            pattern = (
                r"[A-Z]{1,8}(?P<weekno>(0[1-9]|[1-4][0-9]|5[0-3]))"
                r"W(?P<year>[0-9]{2})"
            )
        case "semimonthly":
            pattern = (
                r"[A-Z]{1,8}(?P<occurrence>[12])"
                r"(?P<month>[FGHJKMNQUVXZ])(?P<year>[0-9]{2})"
            )
        case "monthly" | "quarterly":
            pattern = r"[A-Z]{1,8}(?P<month>[FGHJKMNQUVXZ])(?P<year>[0-9]{2})"
        case "yearly":
            pattern = r"[A-Z]{1,8}(?P<year>[0-9]{4})"
        case _:
            raise ValueError(
                f"OutcomeSpaceTimeSeries: frequency: Unexpected value '{frequency}'"
            )
    if match := re.fullmatch(pattern, symbol):
        groups = match.groupdict()

        if "day" in groups and int(groups["day"]) != release_date.day:
            raise ValueError(
                f"ProductMetadata: symbol: Day component of symbol '{symbol}' "
                f"should match TemporalObservation.release_date {release_date}"
            )
        if "month" in groups and MONTH_CODES[groups["month"]] != release_date.month:
            raise ValueError(
                f"ProductMetadata: symbol: Month component of symbol '{symbol}' "
                f"should match TemporalObservation.release_date {release_date}"
            )
        if "year" in groups and int(groups["year"]) % 2000 != release_date.year % 2000:
            raise ValueError(
                f"ProductMetadata: symbol: Year component of symbol '{symbol}' "
                f"should match TemporalObservation.release_date {release_date}"
            )
        if (
            "weekno" in groups
            and int(groups["weekno"]) != release_date.isocalendar().week
        ):
            raise ValueError(
                f"ProductMetadata: symbol: Week number component of symbol '{symbol}' "
                f"should match TemporalObservation.release_date {release_date}"
            )
        if "occurrence" in groups and int(
            groups["occurrence"]
        ) != _semimonthly_occurrence(release_date.day):
            raise ValueError(
                f"ProductMetadata: symbol: Occurrence component of symbol '{symbol}' "
                f"should match TemporalObservation.release_date {release_date}"
            )
    else:
        raise ValueError(
            f"ProductMetadata: symbol: Symbol '{symbol}' should match regexp pattern "
            f"'{pattern}' for OutcomeSpaceTimeSeries.frequency '{frequency}'"
        )


def _semimonthly_occurrence(day_of_month: int) -> int:
    return 1 if day_of_month <= 15 else 2



import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes


class Side(enum.IntEnum):
    """Port of `enum Side` on the ClearingFacet contract."""

    BID = 0
    ASK = 1


class ProductState(enum.IntEnum):
    """Port of `enum ProductState` on the ProductRegistry contract."""

    NOT_EXIST = 0
    PENDING = 1
    LIVE = 2
    TRADEOUT = 3
    FINAL_SETTLEMENT = 4
    EXPIRED = 5


@dataclass
class ClearingConfig:
    """Port of `struct ClearingConfig` on the AdminFacet contract."""

    clearing_fee_rate: int
    max_trading_fee_rate: int


@dataclass
class MarkPriceConfig:
    """Port of `struct MarkPriceConfig` on the AdminFacet contract."""

    mark_price_interval: int


@dataclass
class FinalSettlementConfig:
    """Port of `struct FinalSettlementConfig` on the AdminFacet contract."""

    closeout_fee_rate: int
    closeout_reward_rate: int


@dataclass
class ProductConfig:
    """Port of `struct ProductConfig` on the AdminFacet contract."""

    clearing_payout_ratio: int
    maintenance_deposit_interval: int
    min_product_maintenance_fee: int


@dataclass
class Config:
    """Port of `struct Config` on the AdminFacet contract."""

    clearing_config: ClearingConfig
    mark_price_config: MarkPriceConfig
    final_settlement_config: FinalSettlementConfig
    product_config: ProductConfig


@dataclass
class IntentData:
    """Port of `struct IntentData` on the ClearingFacet contract."""

    nonce: int
    trading_protocol_id: eth_typing.ChecksumAddress
    product_id: hexbytes.HexBytes
    limit_price: int
    quantity: int
    max_trading_fee_rate: int
    good_until: int
    side: Side
    referral: eth_typing.ChecksumAddress


@dataclass
class Intent:
    """Port of `struct Intent` on the ClearingFacet contract."""

    margin_account_id: eth_typing.ChecksumAddress
    intent_account_id: eth_typing.ChecksumAddress
    hash: hexbytes.HexBytes
    data: IntentData
    signature: hexbytes.HexBytes


@dataclass
class Trade:
    """Port of `struct Trade` on the ClearingFacet contract."""

    product_id: hexbytes.HexBytes
    protocol_id: eth_typing.ChecksumAddress
    trade_id: int
    price: int
    timestamp: int
    quantity: int
    accounts: typing.List[eth_typing.ChecksumAddress]
    quantities: typing.List[int]
    fee_rates: typing.List[int]
    intents: typing.List[Intent]


@dataclass
class Settlement:
    """Port of `struct Settlement` on the MarginAccountFacet contract."""

    product_id: hexbytes.HexBytes
    quantity: int
    price: int


@dataclass
class MarginData:
    """Port of `struct MarginData` on the IMarginAccountFacet contract."""

    capital: int
    mae: int
    mmu: int
    pnl: int


@dataclass
class PositionData:
    """Port of `struct PositionData` on the MarginAccountFacet contract."""

    product_id: hexbytes.HexBytes
    quantity: int
    cost_basis: int
    maintenance_margin: int
    pnl: int


@dataclass
class ExpirySpecification:
    """Port of `struct ExpirySpecification` on the ProductRegistryFacet contract."""

    earliest_fsp_submission_time: int
    tradeout_interval: int


@dataclass
class PendingStakeData:
    """Port of `struct PendingStakeData` on the ProductRegistry contract."""

    payout: int
    active_at: int


@dataclass
class BuilderStakeData:
    """Port of `struct BuilderStakeData` on the ProductRegistry contract."""

    payout_remaining: int
    claimable_rewards: int
    pending_stake_data: PendingStakeData


@dataclass
class ProductMetadata:
    """Port of `struct ProductMetadata` on the ProductRegistryFacet contract."""

    builder: eth_typing.ChecksumAddress
    symbol: str
    description: str


@dataclass
class OracleSpecification:
    """Port of `struct OracleSpecification` on the ProductRegistryFacet contract."""

    oracle_address: eth_typing.ChecksumAddress
    fsv_decimals: int
    fsp_alpha: int
    fsp_beta: int
    fsv_calldata: hexbytes.HexBytes


@dataclass
class BaseProduct:
    """Port of `struct BaseProduct` on the ProductRegistryFacet contract."""

    metadata: ProductMetadata
    oracle_spec: OracleSpecification
    collateral_asset: eth_typing.ChecksumAddress
    start_time: int
    point_value: int
    price_decimals: int
    extended_metadata: str


@dataclass
class MarginSpecification:
    """Port of `struct MarginSpecification` on the ProductRegistryFacet contract."""

    imr: int
    mmr: int


@dataclass
class FuturesProductV1:
    """Port of `struct FuturesProductV1` on the ProductRegistryFacet contract."""

    base: BaseProduct
    expiry_spec: ExpirySpecification
    margin_spec: MarginSpecification


@dataclass
class PredictionProductV1:
    """Port of `struct PredictionProductV1` on the ProductRegistryFacet contract."""

    base: BaseProduct
    expiry_spec: ExpirySpecification
    max_price: int
    min_price: int



"""Comprehensive integration tests for IPFSClient."""

import json
from unittest.mock import Mock

import dag_cbor
import pytest
import requests
from requests import Response

from afp import constants
from afp.exceptions import IPFSError, ValidationError
from afp.ipfs import IPFSClient
from afp.schemas import OracleConfig

from .fixtures import (
    make_extended_metadata,
    make_ipfs_block_response,
    make_ipfs_car_response,
    make_oracle_fallback,
    make_outcome_point_time_series,
    make_outcome_space_time_series,
)


def test_upload_extended_metadata__success__returns_cid(monkeypatch):
    """Test successful upload of extended metadata."""
    metadata = make_extended_metadata()

    # We need to mock _load_schema_json and ensure_cids_match to avoid file I/O
    # and CID validation issues in this integration test
    def mock_load_schema(cid):
        # Return the actual schema JSON content that would be loaded
        import json
        import os

        try:
            with open(
                os.path.join(constants.JSON_SCHEMAS_DIRECTORY, f"{cid}.json")
            ) as f:
                return json.load(f)
        except FileNotFoundError:
            # Fallback for test schemas
            return {"$schema": "test", "type": "object"}

    def mock_ensure_match(cid1, cid2):
        # Skip CID validation in this test
        pass

    monkeypatch.setattr(
        "afp.ipfs.IPFSClient._load_schema_json", staticmethod(mock_load_schema)
    )
    monkeypatch.setattr(
        "afp.ipfs.IPFSClient.ensure_cids_match", staticmethod(mock_ensure_match)
    )

    # Encode a sample DAG to get a valid CID
    sample_dag, _ = IPFSClient.encode({"test": "data"})
    root_cid = str(sample_dag)

    fake_response = make_ipfs_car_response(root_cid)
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    result_cid = client.upload_extended_metadata(metadata)

    assert isinstance(result_cid, str)
    assert result_cid.startswith("bafy")

    # Verify CAR upload
    call_kwargs = mock_post.call_args[1]
    assert "files" in call_kwargs
    assert call_kwargs["files"]["file"][2] == "application/vnd.ipld.car"


def test_upload_car__success__returns_root_cid(monkeypatch):
    """Test successful CAR file upload."""

    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    root_cid_str = str(cid)
    fake_response = make_ipfs_car_response(root_cid_str)
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    result_cid = client.upload_car(blocks)  # type: ignore[arg-type]

    assert result_cid == root_cid_str

    # Verify request
    assert mock_post.called
    call_kwargs = mock_post.call_args[1]
    assert "pin-roots" in call_kwargs["params"]
    assert call_kwargs["params"]["pin-roots"] == "true"


def test_upload_car__sends_car_content_type(monkeypatch):
    """Test that CAR upload sends correct content type."""
    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    fake_response = make_ipfs_car_response(str(cid))
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    client.upload_car(blocks)  # type: ignore[arg-type]

    call_kwargs = mock_post.call_args[1]
    assert call_kwargs["files"]["file"][2] == "application/vnd.ipld.car"


def test_upload_car__invalid_response__raises_ipfs_error(monkeypatch):
    """Test that invalid IPFS response raises IPFSError."""
    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    fake_response = Mock(spec=Response)
    fake_response.status_code = 200
    fake_response.json.return_value = {"Invalid": "Structure"}
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="Unexpected IPFS response format"):
        client.upload_car(blocks)  # type: ignore[arg-type]


def test_upload_car__json_decode_error__raises_ipfs_error(monkeypatch):
    """Test that JSON decode error raises IPFSError."""
    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    fake_response = Mock(spec=Response)
    fake_response.status_code = 200
    fake_response.json.side_effect = json.JSONDecodeError("Invalid JSON", "", 0)
    fake_response.text = "Invalid response"
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="Error decoding IPFS response"):
        client.upload_car(blocks)  # type: ignore[arg-type]


def test_download_extended_metadata__success__returns_metadata(monkeypatch):
    """Test successful download and reconstruction of extended metadata."""

    outcome_space = make_outcome_space_time_series()
    outcome_point = make_outcome_point_time_series()
    oracle_config = OracleConfig(
        description="Test oracle",
        project_url="https://example.com",
    )
    oracle_fallback = make_oracle_fallback()

    # Encode all components
    os_cid, os_data = IPFSClient.encode(outcome_space.model_dump(mode="json"))
    op_cid, op_data = IPFSClient.encode(outcome_point.model_dump(mode="json"))
    oc_cid, oc_data = IPFSClient.encode(oracle_config.model_dump(mode="json"))
    of_cid, of_data = IPFSClient.encode(oracle_fallback.model_dump(mode="json"))

    # Create DAG root manually to avoid validation issues
    dag_dict = {
        "outcome_space": {
            "data": str(os_cid),
            "schema": outcome_space.SCHEMA_CID,
        },
        "outcome_point": {
            "data": str(op_cid),
            "schema": outcome_point.SCHEMA_CID,
        },
        "oracle_config": {
            "data": str(oc_cid),
            "schema": oracle_config.SCHEMA_CID,
        },
        "oracle_fallback": {
            "data": str(of_cid),
            "schema": oracle_fallback.SCHEMA_CID,
        },
    }
    dag_cid, dag_data = IPFSClient.encode(dag_dict)

    # Mock IPFS responses
    def mock_post_handler(url, **kwargs):
        if "/block/get" in url:
            arg = kwargs["params"]["arg"]
            response_map = {
                str(dag_cid): dag_data,
                str(os_cid): os_data,
                str(op_cid): op_data,
                str(oc_cid): oc_data,
                str(of_cid): of_data,
            }
            return make_ipfs_block_response(response_map.get(arg, b""))
        return Response()

    mock_post = Mock(side_effect=mock_post_handler)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    result = client.download_extended_metadata(str(dag_cid))

    assert result.outcome_space.description == "Test outcome space"
    assert result.oracle_config.description == "Test oracle"


def test_download_and_validate_block__dag_cbor_codec__decodes_correctly(monkeypatch):
    """Test that DAG-CBOR blocks are decoded correctly."""
    data = {"test": "value", "number": 123}
    cid, encoded = IPFSClient.encode(data)

    fake_response = make_ipfs_block_response(encoded)
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    from afp.dtos import ExtendedMetadataDAG

    client = IPFSClient("http://ipfs.test")
    # We'll get a validation error because the data doesn't match schema,
    # but we can test the codec path by catching it
    try:
        client._download_and_validate_block(str(cid), ExtendedMetadataDAG)
    except Exception:
        pass

    # Verify the request was made
    assert mock_post.called
    assert "/block/get" in mock_post.call_args[0][0]


def test_download_and_validate_block__unsupported_codec__raises_error(monkeypatch):
    """Test that unsupported codec raises IPFSError."""
    # Create a CID with unsupported codec
    import multiformats

    data = b"test data"
    digest = multiformats.multihash.digest(data, "sha2-256")
    # Use raw codec instead of dag-cbor
    cid = multiformats.CID("base32", 1, "raw", digest)

    fake_response = make_ipfs_block_response(data)
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    from afp.dtos import ExtendedMetadataDAG

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="Unsupported codec"):
        client._download_and_validate_block(str(cid), ExtendedMetadataDAG)


def test_encode__returns_cid_and_bytes():
    """Test that encode returns both CID and encoded bytes."""
    data = {"test": "value", "number": 42}

    cid, encoded = IPFSClient.encode(data)

    assert isinstance(cid, object)  # multiformats.CID
    assert isinstance(encoded, bytes)
    assert len(encoded) > 0


def test_encode__cid_is_deterministic():
    """Test that same input produces same CID."""
    data = {"key": "value"}

    cid1, _ = IPFSClient.encode(data)
    cid2, _ = IPFSClient.encode(data)

    assert str(cid1) == str(cid2)


def test_encode__uses_dag_cbor_codec():
    """Test that encoding uses dag-cbor codec."""
    data = {"test": "data"}
    cid, _ = IPFSClient.encode(data)

    import multiformats

    decoded_cid = multiformats.CID.decode(str(cid))
    assert decoded_cid.codec.name == "dag-cbor"


def test_encode__uses_sha256_hash():
    """Test that encoding uses SHA-256 hash function."""
    data = {"test": "data"}
    cid, _ = IPFSClient.encode(data)

    import multiformats

    decoded_cid = multiformats.CID.decode(str(cid))
    # Verify the digest is a bytes object (multihash)
    assert isinstance(decoded_cid.digest, bytes)
    # SHA-256 produces 32-byte hash, plus 2-byte multihash header (code + length)
    assert len(decoded_cid.digest) == 34


def test_encode__uses_base32_encoding():
    """Test that CID uses base32 encoding."""
    data = {"test": "data"}
    cid, _ = IPFSClient.encode(data)

    # Base32 CIDs start with 'b'
    cid_str = str(cid)
    assert cid_str.startswith("b")
    assert cid_str.startswith("bafy")  # dag-cbor with base32


def test_encode__can_decode_encoded_data():
    """Test that encoded data can be decoded back."""
    data = {"test": "value", "nested": {"key": 123}}
    _, encoded = IPFSClient.encode(data)

    decoded = dag_cbor.decode(encoded)
    assert decoded == data


def test_ensure_cids_match__matching_cids__no_error():
    """Test that matching CIDs don't raise an error."""
    cid1 = "bafyreiabc123"
    cid2 = "bafyreiabc123"

    # Should not raise
    IPFSClient.ensure_cids_match(cid1, cid2)


def test_ensure_cids_match__mismatched_cids__raises_ipfs_error():
    """Test that mismatched CIDs raise IPFSError."""
    cid1 = "bafyreiabc123"
    cid2 = "bafyreiabc456"

    with pytest.raises(IPFSError, match="Mismatch between computed CID"):
        IPFSClient.ensure_cids_match(cid1, cid2)


def test_find_model_by_schema_cid__valid_cid__returns_model():
    """Test that valid schema CID returns correct model."""
    from afp.schemas import OutcomeSpace

    model = IPFSClient._find_model_by_schema_cid(
        constants.schema_cids.OUTCOME_SPACE_V020
    )

    assert model == OutcomeSpace


def test_find_model_by_schema_cid__invalid_cid__raises_validation_error():
    """Test that invalid schema CID raises ValidationError."""
    with pytest.raises(ValidationError, match="Unsupported schema CID"):
        IPFSClient._find_model_by_schema_cid("bafyreiinvalidcid123")


def test_load_schema_json__valid_cid__returns_dict():
    """Test that loading schema JSON returns a dictionary."""
    result = IPFSClient._load_schema_json(constants.schema_cids.OUTCOME_SPACE_V020)

    assert isinstance(result, dict)
    assert "$schema" in result or "type" in result


def test_load_schema_json__contains_expected_fields():
    """Test that loaded schema has expected structure."""
    schema = IPFSClient._load_schema_json(constants.schema_cids.OUTCOME_SPACE_V020)

    assert isinstance(schema, dict)
    # JSON schemas typically have these fields
    assert "type" in schema or "properties" in schema or "$schema" in schema


def test_init__with_api_key__sets_authorization_header(monkeypatch):
    """Test that API key is included in authorization header."""
    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    fake_response = make_ipfs_car_response(str(cid))
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test", api_key="test-api-key")
    client.upload_car(blocks)  # type: ignore[arg-type]

    call_kwargs = mock_post.call_args[1]
    assert "headers" in call_kwargs
    assert call_kwargs["headers"]["Authorization"] == "Bearer test-api-key"


def test_init__without_api_key__no_authorization_header(monkeypatch):
    """Test that no API key means no authorization header."""
    cid, data = IPFSClient.encode({"test": "data"})
    blocks = [(cid, data)]

    fake_response = make_ipfs_car_response(str(cid))
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test", api_key=None)
    client.upload_car(blocks)  # type: ignore[arg-type]

    call_kwargs = mock_post.call_args[1]
    assert "headers" in call_kwargs
    assert "Authorization" not in call_kwargs["headers"]


def test_send_client_request__timeout__raises_ipfs_error(monkeypatch):
    """Test that request timeout raises IPFSError."""
    mock_post = Mock(side_effect=requests.exceptions.Timeout("Connection timeout"))
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="timed out after"):
        client._send_client_request("/api/v0/block/get", params={"arg": "test"})


def test_send_client_request__connection_error__raises_ipfs_error(monkeypatch):
    """Test that connection error raises IPFSError."""
    mock_post = Mock(
        side_effect=requests.exceptions.ConnectionError("Connection refused")
    )
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="Failed to send request"):
        client._send_client_request("/api/v0/block/get", params={"arg": "test"})


def test_send_client_request__http_error__raises_ipfs_error(monkeypatch):
    """Test that HTTP error raises IPFSError."""
    fake_response = Mock(spec=Response)
    fake_response.status_code = 500
    fake_response.raise_for_status.side_effect = requests.exceptions.HTTPError(
        response=fake_response
    )
    mock_post = Mock(return_value=fake_response)
    monkeypatch.setattr(requests, "post", mock_post)

    client = IPFSClient("http://ipfs.test")
    with pytest.raises(IPFSError, match="HTTP 500"):
        client._send_client_request("/api/v0/block/get", params={"arg": "test"})



from dataclasses import dataclass

from eth_typing.evm import ChecksumAddress

from .auth import Authenticator


@dataclass(frozen=True)
class Config:
    authenticator: Authenticator | None

    # IPFS client parameters
    ipfs_api_url: str
    ipfs_api_key: str | None

    # Venue parameters
    exchange_url: str

    # Blockchain parameters
    rpc_url: str | None
    chain_id: int
    gas_limit: int | None
    max_fee_per_gas: int | None
    max_priority_fee_per_gas: int | None
    timeout_seconds: int

    # Clearing System parameters
    clearing_diamond_address: ChecksumAddress
    margin_account_registry_address: ChecksumAddress
    oracle_provider_address: ChecksumAddress
    product_registry_address: ChecksumAddress
    system_viewer_address: ChecksumAddress



from typing import Any, cast

from eth_abi.packed import encode_packed
from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

from . import constants
from .bindings import Side as OnChainSide
from .schemas import IntentData, OrderSide


ORDER_SIDE_MAPPING: dict[OrderSide, int] = {
    OrderSide.BID: OnChainSide.BID.value,
    OrderSide.ASK: OnChainSide.ASK.value,
}


def generate_intent_hash(
    intent_data: IntentData,
    margin_account_id: ChecksumAddress,
    intent_account_id: ChecksumAddress,
    tick_size: int,
) -> HexBytes:
    types = [
        "address",  # margin_account_id
        "address",  # intent_account_id
        "uint256",  # nonce
        "address",  # trading_protocol_id
        "bytes32",  # product_id
        "int256",  # limit_price
        "uint256",  # quantity
        "uint32",  # max_trading_fee_rate
        "uint256",  # good_until_time
        "uint8",  # side
        "address",  # referral
    ]
    values: list[Any] = [
        margin_account_id,
        intent_account_id,
        intent_data.nonce,
        cast(ChecksumAddress, intent_data.trading_protocol_id),
        HexBytes(intent_data.product_id),
        int(intent_data.limit_price * 10**tick_size),
        intent_data.quantity,
        int(intent_data.max_trading_fee_rate * constants.FEE_RATE_MULTIPLIER),
        int(intent_data.good_until_time.timestamp()),
        ORDER_SIDE_MAPPING[intent_data.side],
        cast(ChecksumAddress, intent_data.referral),
    ]
    return Web3.keccak(encode_packed(types, values))


def generate_order_cancellation_hash(nonce: int, intent_hash: str) -> HexBytes:
    types = ["uint256", "bytes32"]
    values: list[Any] = [nonce, HexBytes(intent_hash)]
    return Web3.keccak(encode_packed(types, values))


def generate_product_id(builder_address: ChecksumAddress, symbol: str) -> HexBytes:
    types = ["address", "string"]
    values: list[Any] = [builder_address, symbol]
    return Web3.keccak(encode_packed(types, values))



"""MarginAccountFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract

from .types import Settlement, MarginData, PositionData


class MarginAccountFacet:
    """MarginAccountFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarginAccountFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def PositionUpdated(self) -> contract.ContractEvent:
        """Binding for `event PositionUpdated` on the MarginAccountFacet contract."""
        return self._contract.events.PositionUpdated

    def authorize(
        self,
        owner: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `authorize` on the MarginAccountFacet contract.

        Parameters
        ----------
        owner : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.authorize(
            owner,
            intent_account,
        )

    def collateral(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateral` on the MarginAccountFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateral().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def deposit(
        self,
        owner: eth_typing.ChecksumAddress,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `deposit` on the MarginAccountFacet contract.

        Parameters
        ----------
        owner : eth_typing.ChecksumAddress
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.deposit(
            owner,
            amount,
        )

    def estimate_additional_margin(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        settlement: Settlement,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `estimateAdditionalMargin` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        settlement : Settlement
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.estimateAdditionalMargin(
            margin_account_id,
            (settlement.product_id, settlement.quantity, settlement.price),
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def is_authorized(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> bool:
        """Binding for `isAuthorized` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.isAuthorized(
            margin_account_id,
            intent_account,
        ).call(block_identifier=block_identifier)
        return bool(return_value)

    def mae_check(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        settlement: Settlement,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> bool:
        """Binding for `maeCheck` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        settlement : Settlement
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.maeCheck(
            margin_account_id,
            (settlement.product_id, settlement.quantity, settlement.price),
        ).call(block_identifier=block_identifier)
        return bool(return_value)

    def margin_data(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> MarginData:
        """Binding for `marginData` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        MarginData
        """
        return_value = self._contract.functions.marginData(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return MarginData(
            int(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
        )

    def position(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> PositionData:
        """Binding for `position` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        PositionData
        """
        return_value = self._contract.functions.position(
            margin_account_id,
            product_id,
        ).call(block_identifier=block_identifier)
        return PositionData(
            hexbytes.HexBytes(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            int(return_value[4]),
        )

    def products_by_margin_account(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[hexbytes.HexBytes]:
        """Binding for `products` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[hexbytes.HexBytes]
        """
        return_value = self._contract.functions.products(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return [
            hexbytes.HexBytes(return_value_elem) for return_value_elem in return_value
        ]

    def revoke(
        self,
        owner: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `revoke` on the MarginAccountFacet contract.

        Parameters
        ----------
        owner : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.revoke(
            owner,
            intent_account,
        )

    def state_after_settlement(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        settlement: Settlement,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, int]:
        """Binding for `stateAfterSettlement` on the MarginAccountFacet contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        settlement : Settlement
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        int
        """
        return_value = self._contract.functions.stateAfterSettlement(
            margin_account_id,
            (settlement.product_id, settlement.quantity, settlement.price),
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            int(return_value[1]),
        )

    def withdraw(
        self,
        owner: eth_typing.ChecksumAddress,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `withdraw` on the MarginAccountFacet contract.

        Parameters
        ----------
        owner : eth_typing.ChecksumAddress
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.withdraw(
            owner,
            amount,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "authorize",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address"},
                {"name": "intentAccount", "type": "address", "internalType": "address"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "collateral",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "deposit",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address"},
                {"name": "amount", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "estimateAdditionalMargin",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "settlement",
                    "type": "tuple",
                    "internalType": "struct Settlement",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {"name": "price", "type": "int256", "internalType": "int256"},
                    ],
                },
            ],
            "outputs": [
                {
                    "name": "additionalMargin",
                    "type": "uint256",
                    "internalType": "uint256",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "isAuthorized",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "intentAccount", "type": "address", "internalType": "address"},
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "maeCheck",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "settlement",
                    "type": "tuple",
                    "internalType": "struct Settlement",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {"name": "price", "type": "int256", "internalType": "int256"},
                    ],
                },
            ],
            "outputs": [
                {"name": "checkPassed", "type": "bool", "internalType": "bool"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "marginData",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct IMarginAccountFacet.MarginData",
                    "components": [
                        {"name": "capital", "type": "int256", "internalType": "int256"},
                        {"name": "mae", "type": "int256", "internalType": "int256"},
                        {"name": "mmu", "type": "uint256", "internalType": "uint256"},
                        {"name": "pnl", "type": "int256", "internalType": "int256"},
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "position",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct PositionData",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "costBasis",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "maintenanceMargin",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {"name": "pnl", "type": "int256", "internalType": "int256"},
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "products",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "bytes32[]", "internalType": "bytes32[]"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "revoke",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address"},
                {"name": "intentAccount", "type": "address", "internalType": "address"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "stateAfterSettlement",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "settlement",
                    "type": "tuple",
                    "internalType": "struct Settlement",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {"name": "price", "type": "int256", "internalType": "int256"},
                    ],
                },
            ],
            "outputs": [
                {"name": "maeAfter", "type": "int256", "internalType": "int256"},
                {"name": "mmuAfter", "type": "uint256", "internalType": "uint256"},
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "withdraw",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address"},
                {"name": "amount", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "event",
            "name": "PositionUpdated",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "positionId",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "costBasis",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "price",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "quantity",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {"type": "error", "name": "EVWMA_NotInitialized", "inputs": []},
        {
            "type": "error",
            "name": "InsufficientBalance",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"},
                {"name": "balance", "type": "uint256", "internalType": "uint256"},
                {"name": "required", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidFieldAccess",
            "inputs": [
                {
                    "name": "productType",
                    "type": "uint8",
                    "internalType": "enum ProductType",
                },
                {"name": "field", "type": "string", "internalType": "string"},
            ],
        },
        {
            "type": "error",
            "name": "NotFound",
            "inputs": [
                {"name": "parameter", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv18_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
                {"name": "denominator", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Div_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Div_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp2_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Mul_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Mul_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedIntToUint",
            "inputs": [{"name": "value", "type": "int256", "internalType": "int256"}],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedUintToInt",
            "inputs": [{"name": "value", "type": "uint256", "internalType": "uint256"}],
        },
        {
            "type": "error",
            "name": "Unauthorized",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
    ],
)



import os
from datetime import datetime, timedelta
from decimal import Decimal
from pprint import pprint
from time import sleep

import afp


AUTONITY_RPC_URL = "https://bakerloo.autonity-apis.com"
TRADER1_PRIVATE_KEY = os.environ["TRADER1_PRIVATE_KEY"]
TRADER2_PRIVATE_KEY = os.environ["TRADER2_PRIVATE_KEY"]

PRODUCT_ID = "0xf82118deb932a8649d519d8d34e7f7b278a44bdb3f2663f6049aaea6ee33b211"
COLLATERAL_ASSET = "0x15d7B75e16162C36d86c971508c0d6eF2D7F131D"


def main():
    app = afp.AFP(rpc_url=AUTONITY_RPC_URL)

    # Trader #1 adds collateral to their margin account
    margin_account = app.MarginAccount(afp.PrivateKeyAuthenticator(TRADER1_PRIVATE_KEY))
    margin_account.deposit(COLLATERAL_ASSET, Decimal("100"))

    # Trader #1 submits bid
    trading = app.Trading(afp.PrivateKeyAuthenticator(TRADER1_PRIVATE_KEY))

    product = trading.product(PRODUCT_ID)
    pprint(product.model_dump())

    bid_intent = trading.create_intent(
        product=product,
        side="bid",
        limit_price=Decimal("1.23"),
        quantity=2,
        max_trading_fee_rate=Decimal("0.1"),
        good_until_time=datetime.now() + timedelta(hours=1),
    )
    pprint(bid_intent.model_dump())

    bid_order = trading.submit_limit_order(bid_intent)
    pprint(bid_order.model_dump())

    # Trader #2 submits ask
    trading = app.Trading(afp.PrivateKeyAuthenticator(TRADER2_PRIVATE_KEY))

    ask_intent = trading.create_intent(
        product=product,
        side="ask",
        limit_price=Decimal("1.23"),
        quantity=2,
        max_trading_fee_rate=Decimal("0.1"),
        good_until_time=datetime.now() + timedelta(hours=1),
    )
    pprint(ask_intent.model_dump())

    ask_order = trading.submit_limit_order(ask_intent)
    pprint(ask_order.model_dump())

    # Check order fills
    sleep(0.5)
    fills = trading.order_fills(product_id=product.id)
    pprint([fill.model_dump() for fill in fills])


if __name__ == "__main__":
    main()



import secrets
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Generator, Iterable

from web3 import Web3
from web3.constants import CHECKSUM_ADDRESSS_ZERO

from .. import hashing, validators
from ..constants import DEFAULT_BATCH_SIZE
from ..decorators import refresh_token_on_expiry
from ..dtos import ExchangeProductFilter, OrderFilter, OrderFillFilter, OrderSubmission
from ..enums import OrderSide, OrderState, OrderType, TradeState
from ..schemas import (
    ExchangeProduct,
    Intent,
    IntentData,
    MarketDepthData,
    OHLCVItem,
    Order,
    OrderCancellationData,
    OrderFill,
)
from .base import ExchangeAPI


class Trading(ExchangeAPI):
    """API for trading in the AutEx exchange."""

    @staticmethod
    def _generate_nonce() -> int:
        return secrets.randbelow(2**31)  # Postgres integer range

    def create_intent(
        self,
        *,
        product: ExchangeProduct,
        side: str,
        limit_price: Decimal,
        quantity: int,
        max_trading_fee_rate: Decimal,
        good_until_time: datetime,
        margin_account_id: str | None = None,
        referral: str | None = None,
        rounding: str | None = None,
    ) -> Intent:
        """Creates an intent with the given intent data, generates its hash and signs it
        with the configured account's private key.

        The intent account's address is derived from the private key. The intent account
        is assumed to be the same as the margin account if the margin account ID is not
        specified.

        The limit price must have at most as many fractional digits as the product's
        tick size, or if `rounding` is specified then the limit price is rounded to the
        appropriate number of fractional digits. `rounding` may be one of
        `ROUND_CEILING`, `ROUND_FLOOR`, `ROUND_UP`, `ROUND_DOWN`, `ROUND_HALF_UP`,
        `ROUND_HALF_DOWN`, `ROUND_HALF_EVEN` and `ROUND_05UP`; see the rounding modes of
        the `decimal` module of the Python Standard Library.

        Parameters
        ----------
        product : afp.schemas.ExchangeProduct
        side : str
            One of `BID` and `ASK`.
        limit_price : decimal.Decimal
        quantity : decimal.Decimal
        max_trading_fee_rate : decimal.Decimal
        good_until_time : datetime.datetime
        margin_account_id : str, optional
        referral : str, optional
        rounding : str, optional
            A rounding mode of the `decimal` module or `None` for no rounding.

        Returns
        -------
        afp.schemas.Intent
        """
        margin_account_id = (
            validators.validate_address(margin_account_id)
            if margin_account_id is not None
            else self._authenticator.address
        )

        intent_data = IntentData(
            trading_protocol_id=self._trading_protocol_id,
            product_id=product.id,
            limit_price=validators.validate_limit_price(
                Decimal(limit_price),
                product.min_price,
                product.max_price,
                product.tick_size,
                rounding,
            ),
            quantity=quantity,
            max_trading_fee_rate=max_trading_fee_rate,
            side=OrderSide(side.upper()),
            good_until_time=good_until_time,
            nonce=self._generate_nonce(),
            referral=(referral if referral is not None else CHECKSUM_ADDRESSS_ZERO),
        )
        intent_hash = hashing.generate_intent_hash(
            intent_data=intent_data,
            margin_account_id=margin_account_id,
            intent_account_id=self._authenticator.address,
            tick_size=product.tick_size,
        )
        signature = self._authenticator.sign_message(intent_hash)
        return Intent(
            hash=Web3.to_hex(intent_hash),
            margin_account_id=margin_account_id,
            intent_account_id=self._authenticator.address,
            signature=Web3.to_hex(signature),
            data=intent_data,
        )

    @refresh_token_on_expiry
    def submit_limit_order(self, intent: Intent) -> Order:
        """Sends an intent expressing a limit order to the exchange.

        Parameters
        ----------
        intent : afp.schemas.Intent

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.RateLimitExceeded
            If the exchange rejects the order because of too many requests from the
            authenticated account.
        """
        submission = OrderSubmission(
            type=OrderType.LIMIT_ORDER,
            intent=intent,
        )
        return self._exchange.submit_order(submission)

    @refresh_token_on_expiry
    def submit_cancel_order(self, intent_hash: str) -> Order:
        """Sends a cancellation order to the exchange.

        Parameters
        ----------
        intent_hash : str

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.NotFoundError
            If the exchange rejects the cancellation because the intent does not exist.
        afp.exceptions.RateLimitExceeded
            If the exchange rejects the cancellation because of too many requests from
            the authenticated account.
        afp.exceptions.ValidationError
            If the exchange rejects the cancellation because it is invalid.
        """
        nonce = self._generate_nonce()
        cancellation_hash = hashing.generate_order_cancellation_hash(nonce, intent_hash)
        signature = self._authenticator.sign_message(cancellation_hash)
        cancellation_data = OrderCancellationData(
            intent_hash=intent_hash,
            nonce=nonce,
            intent_account_id=self._authenticator.address,
            signature=Web3.to_hex(signature),
        )
        submission = OrderSubmission(
            type=OrderType.CANCEL_ORDER,
            cancellation_data=cancellation_data,
        )
        return self._exchange.submit_order(submission)

    def products(
        self,
        batch: int = 1,
        batch_size: int = DEFAULT_BATCH_SIZE,
        newest_first: bool = True,
    ) -> list[ExchangeProduct]:
        """Retrieves the products approved for trading on the exchange.

        If there are more than `batch_size` number of products then they can be queried
        in batches.

        Parameters
        ----------
        batch : int, optional
            1-based index of the batch of products.
        batch_size : int, optional
            The maximum number of products in one batch.
        newest_first : bool, optional
            Whether to sort products in descending or ascending order by creation time.

        Returns
        -------
        list of afp.schemas.ExchangeProduct
        """
        filter = ExchangeProductFilter(
            batch=batch, batch_size=batch_size, newest_first=newest_first
        )
        return self._exchange.get_approved_products(filter)

    def product(self, product_id: str) -> ExchangeProduct:
        """Retrieves a product for trading by its ID.

        Parameters
        ----------
        product_id : str

        Returns
        -------
        afp.schemas.ExchangeProduct

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        return self._exchange.get_product_by_id(value)

    @refresh_token_on_expiry
    def order(self, order_id: str) -> Order:
        """Retrieves an order by its ID from the orders that have been submitted by the
        authenticated account.

        Parameters
        ----------
        order_id : str

        Returns
        -------
        afp.schemas.Order

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such order exists.
        """
        value = validators.validate_hexstr32(order_id)
        return self._exchange.get_order_by_id(value)

    @refresh_token_on_expiry
    def orders(
        self,
        *,
        product_id: str | None = None,
        intent_account_id: str | None = None,
        type_: str | None = None,
        states: Iterable[str] = (),
        side: str | None = None,
        start: datetime | None = None,
        end: datetime | None = None,
        batch: int = 1,
        batch_size: int = DEFAULT_BATCH_SIZE,
        newest_first: bool = True,
    ) -> list[Order]:
        """Retrieves the authenticated account's orders that match the given parameters.

        If there are more than `batch_size` number of orders then they can be queried
        in batches.

        Parameters
        ----------
        product_id : str, optional
        intent_account_id : str, optional
            Defaults to the address of the authenticated account.
        type_ : str, optional
            One of `LIMIT_ORDER` and `CANCEL_ORDER`.
        states : iterable of str
            Any of `RECEIVED`, `PENDING`, `OPEN`, `COMPLETED` and `REJECTED`.
        side : str, optional
            One of `BID` and `ASK`.
        start : datetime.datetime, optional
        end : datetime.datetime, optional
        batch : int, optional
            1-based index of the batch of orders.
        batch_size : int, optional
            The maximum number of orders in one batch.
        newest_first : bool, optional
            Whether to sort orders in descending or ascending order by creation time.

        Returns
        -------
        list of afp.schemas.OrderFill
        """
        if intent_account_id is None:
            intent_account_id = self._authenticator.address

        filter = OrderFilter(
            intent_account_id=intent_account_id,
            product_id=product_id,
            type=None if type_ is None else OrderType(type_.upper()),
            states=[OrderState(state.upper()) for state in states],
            side=None if side is None else OrderSide(side.upper()),
            start=start,
            end=end,
            batch=batch,
            batch_size=batch_size,
            newest_first=newest_first,
        )
        return self._exchange.get_orders(filter)

    @refresh_token_on_expiry
    def order_fills(
        self,
        *,
        product_id: str | None = None,
        intent_account_id: str | None = None,
        intent_hash: str | None = None,
        start: datetime | None = None,
        end: datetime | None = None,
        trade_states: Iterable[str] = (),
        batch: int = 1,
        batch_size: int = DEFAULT_BATCH_SIZE,
        newest_first: bool = True,
    ) -> list[OrderFill]:
        """Retrieves the authenticated account's order fills that match the given
        parameters.

        If there are more than `batch_size` number of order fills then they can be
        queried in batches.

        Parameters
        ----------
        product_id : str, optional
        intent_account_id : str, optional
            Defaults to the address of the authenticated account.
        intent_hash : str, optional
        start : datetime.datetime, optional
        end : datetime.datetime, optional
        trade_states : iterable of str
            Any of `PENDING`, `CLEARED` and `REJECTED`.
        batch : int, optional
            1-based index of the batch of order fills.
        batch_size : int, optional
            The maximum number of order fills in one batch.
        newest_first : bool, optional
            Whether to sort order fills in descending or ascending order by creation time.

        Returns
        -------
        list of afp.schemas.OrderFill
        """
        if intent_account_id is None:
            intent_account_id = self._authenticator.address

        filter = OrderFillFilter(
            intent_account_id=intent_account_id,
            product_id=product_id,
            intent_hash=intent_hash,
            start=start,
            end=end,
            trade_states=[TradeState(state.upper()) for state in trade_states],
            batch=batch,
            batch_size=batch_size,
            newest_first=newest_first,
        )
        return self._exchange.get_order_fills(filter)

    @refresh_token_on_expiry
    def iter_order_fills(
        self,
        *,
        product_id: str | None = None,
        intent_account_id: str | None = None,
        intent_hash: str | None = None,
        trade_states: Iterable[str] = ("PENDING",),
    ) -> Generator[OrderFill, None, None]:
        """Subscribes to the authenticated account's new order fills that match the
        given parameters.

        Returns a generator that yields order fills as they are published by the
        exchange.

        If `trade_states` includes `PENDING` (the default value) then a new order fill
        is yielded as soon as there is a match in the order book, before the trade is
        submitted to clearing.

        If `trade_states` is empty or more than one trade state is specified then
        updates to order fills are yielded at every state transition.

        Parameters
        ----------
        product_id : str, optional
        intent_account_id : str, optional
            Defaults to the address of the authenticated account.
        intent_hash : str, optional
        trade_states: iterable of str
            Any of `PENDING`, `CLEARED` and `REJECTED`.

        Yields
        -------
        afp.schemas.OrderFill
        """
        if intent_account_id is None:
            intent_account_id = self._authenticator.address

        filter = OrderFillFilter(
            intent_account_id=intent_account_id,
            product_id=product_id,
            intent_hash=intent_hash,
            start=None,
            end=None,
            trade_states=[TradeState(state.upper()) for state in trade_states],
            batch=None,
            batch_size=None,
            newest_first=None,
        )
        yield from self._exchange.iter_order_fills(filter)

    def market_depth(self, product_id: str) -> MarketDepthData:
        """Retrieves the depth of market for the given product.

        Parameters
        ----------
        product_id : str

        Returns
        -------
        afp.schemas.MarketDepthData

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        return self._exchange.get_market_depth_data(value)

    def iter_market_depth(
        self, product_id: str
    ) -> Generator[MarketDepthData, None, None]:
        """Subscribes to updates of the depth of market for the given product.

        Returns a generator that yields the updated market depth data as it is published
        by the exhange.

        Parameters
        ----------
        product_id : str

        Yields
        -------
        afp.schemas.MarketDepthData

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        value = validators.validate_hexstr32(product_id)
        yield from self._exchange.iter_market_depth_data(value)

    def ohlcv(
        self,
        product_id: str,
        start: datetime | None = None,
        interval: timedelta = timedelta(minutes=5),
    ) -> list[OHLCVItem]:
        """Retrieves Open-High-Low-Close-Volume time series data for the given product.

        Parameters
        ----------
        product_id : str
        start : datetime
            Defaults to 1 day ago.
        interval : timedelta
            The distance between 2 data points. Gets rounded to a multiple of 5 seconds.
            Defaults to 5 minutes.

        Returns
        -------
        list of afp.schemas.OHLCVItem

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        if start is None:
            start = datetime.now() - timedelta(days=1)

        product_id = validators.validate_hexstr32(product_id)
        start_timestamp = int(start.timestamp())
        interval_secs = int(validators.validate_timedelta(interval).total_seconds())
        return self._exchange.get_time_series_data(
            product_id, start_timestamp, interval_secs
        )

    def iter_ohlcv(
        self, product_id: str, interval: timedelta = timedelta(seconds=5)
    ) -> Generator[OHLCVItem, None, None]:
        """Subscribes to Open-High-Low-Close-Volume time series data updates for the
        given product.

        Returns a generator that yields OHLCV data points as they are published
        by the exhange.

        Parameters
        ----------
        product_id : str
        interval : timedelta
            The distance between 2 data points. Gets rounded to a multiple of 5 seconds.
            Defaults to 5 seconds.

        Yields
        -------
        afp.schemas.OHLCVItem

        Raises
        ------
        afp.exceptions.NotFoundError
            If no such product exists.
        """
        product_id = validators.validate_hexstr32(product_id)
        start_timestamp = int(datetime.now().timestamp())
        interval_secs = int(validators.validate_timedelta(interval).total_seconds())
        yield from self._exchange.iter_time_series_data(
            product_id, start_timestamp, interval_secs
        )



import json
import re
from typing import Any, Generator

import requests
from requests import Response, Session

from . import constants
from .dtos import (
    ExchangeParameters,
    ExchangeProductFilter,
    ExchangeProductListingSubmission,
    ExchangeProductUpdateSubmission,
    LoginSubmission,
    OrderFilter,
    OrderFillFilter,
    OrderSubmission,
)
from .exceptions import (
    AuthenticationError,
    AuthorizationError,
    ExchangeError,
    NotFoundError,
    RateLimitExceeded,
    ValidationError,
)
from .schemas import (
    ExchangeProduct,
    MarketDepthData,
    OHLCVItem,
    Order,
    OrderFill,
)


class ExchangeClient:
    _base_url: str
    _session: Session

    def __init__(self, base_url: str):
        self._base_url = re.sub(r"/$", "", base_url)
        self._session = Session()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(base_url={self._base_url})"

    # POST /nonce
    def generate_login_nonce(self) -> str:
        response = self._send_request("GET", "/nonce")
        return response.json()["message"]

    # POST /login
    def login(self, login_submission: LoginSubmission) -> ExchangeParameters:
        response = self._send_request(
            "POST", "/login", data=login_submission.model_dump_json()
        )
        return ExchangeParameters(**response.json())

    # GET /products
    def get_approved_products(
        self, filter: ExchangeProductFilter
    ) -> list[ExchangeProduct]:
        response = self._send_request(
            "GET", "/products", params=filter.model_dump(exclude_none=True)
        )
        return [ExchangeProduct(**item) for item in response.json()["products"]]

    # GET /products/{product_id}
    def get_product_by_id(self, product_id: str) -> ExchangeProduct:
        response = self._send_request("GET", f"/products/{product_id}")
        return ExchangeProduct(**response.json())

    # POST /products
    def list_product(
        self, listing_submission: ExchangeProductListingSubmission
    ) -> None:
        self._send_request(
            "POST", "/products", data=listing_submission.model_dump_json()
        )

    # PATCH /products
    def update_product_listing(
        self, product_id: str, update_submission: ExchangeProductUpdateSubmission
    ) -> None:
        self._send_request(
            "PATCH",
            f"/products/{product_id}",
            data=update_submission.model_dump_json(),
        )

    # POST /orders
    def submit_order(self, order_submission: OrderSubmission) -> Order:
        response = self._send_request(
            "POST", "/orders", data=order_submission.model_dump_json()
        )
        return Order(**response.json())

    # GET /orders
    def get_orders(self, filter: OrderFilter) -> list[Order]:
        response = self._send_request(
            "GET", "/orders", params=filter.model_dump(exclude_none=True)
        )
        return [Order(**item) for item in response.json()["orders"]]

    # GET /orders/{order_id}
    def get_order_by_id(self, order_id: str) -> Order:
        response = self._send_request("GET", f"/orders/{order_id}")
        return Order(**response.json())

    # GET /order-fills
    def get_order_fills(self, filter: OrderFillFilter) -> list[OrderFill]:
        response = self._send_request(
            "GET", "/order-fills", params=filter.model_dump(exclude_none=True)
        )
        return [OrderFill(**item) for item in response.json()["orderFills"]]

    # GET /stream/order-fills
    def iter_order_fills(
        self, filter: OrderFillFilter
    ) -> Generator[OrderFill, None, None]:
        response = self._send_request(
            "GET",
            "/stream/order-fills",
            params=filter.model_dump(exclude_none=True),
            stream=True,
        )
        for line in response.iter_lines():
            yield OrderFill.model_validate_json(line)

    # GET /market-depth/{product_id}
    def get_market_depth_data(self, product_id: str) -> MarketDepthData:
        response = self._send_request("GET", f"/market-depth/{product_id}")
        return MarketDepthData(**response.json())

    # GET /stream/market-depth/{product_id}
    def iter_market_depth_data(
        self, product_id: str
    ) -> Generator[MarketDepthData, None, None]:
        response = self._send_request(
            "GET", f"/stream/market-depth/{product_id}", stream=True
        )
        for line in response.iter_lines():
            yield MarketDepthData.model_validate_json(line)

    # GET /time-series/{product_id}
    def get_time_series_data(
        self, product_id: str, start: int, interval: int
    ) -> list[OHLCVItem]:
        response = self._send_request(
            "GET",
            f"/time-series/{product_id}",
            params=dict(start=start, interval=interval),
        )
        return [OHLCVItem(**item) for item in response.json()["data"]]

    # GET /stream/time-series/{product_id}
    def iter_time_series_data(
        self, product_id: str, start: int, interval: int
    ) -> Generator[OHLCVItem, None, None]:
        response = self._send_request(
            "GET",
            f"/stream/time-series/{product_id}",
            params=dict(start=start, interval=interval),
            stream=True,
        )
        for line in response.iter_lines():
            yield OHLCVItem.model_validate_json(line)

    def _send_request(
        self,
        method: str,
        endpoint: str,
        *,
        stream: bool = False,
        api_version: int = constants.DEFAULT_EXCHANGE_API_VERSION,
        **kwargs: Any,
    ) -> Response:
        kwargs["headers"] = {
            "Content-Type": "application/json",
            "Accept": "application/x-ndjson" if stream else "application/json",
            "User-Agent": constants.USER_AGENT,
        }

        try:
            response = self._session.request(
                method,
                f"{self._base_url}/v{api_version}{endpoint}",
                stream=stream,
                **kwargs,
            )
        except requests.exceptions.RequestException as request_exception:
            raise ExchangeError(
                "Failed to send request to the exchange"
            ) from request_exception
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            if http_error.response.status_code == requests.codes.UNAUTHORIZED:
                raise AuthenticationError(http_error) from http_error
            if http_error.response.status_code == requests.codes.FORBIDDEN:
                raise AuthorizationError(http_error) from http_error
            if http_error.response.status_code == requests.codes.NOT_FOUND:
                raise NotFoundError(http_error) from http_error
            if http_error.response.status_code == requests.codes.TOO_MANY_REQUESTS:
                raise RateLimitExceeded(http_error) from http_error
            if http_error.response.status_code == requests.codes.BAD_REQUEST:
                try:
                    reason = response.json()["detail"]
                except (json.JSONDecodeError, KeyError):
                    reason = http_error
                raise ValidationError(reason) from http_error
            if http_error.response.status_code == requests.codes.UNPROCESSABLE:
                try:
                    reason = ", ".join(err["msg"] for err in response.json()["detail"])
                except (json.JSONDecodeError, KeyError, TypeError):
                    reason = http_error
                raise ValidationError(reason) from http_error

            raise ExchangeError(http_error) from http_error

        return response



"""Internal schemas."""

from decimal import Decimal
from typing import Annotated, Literal

from pydantic import AfterValidator, ConfigDict, Field, computed_field

from . import validators
from .enums import ListingState, OrderSide, OrderState, OrderType, TradeState
from .schemas import (
    Intent,
    OracleConfig,
    OracleConfigPrototype1,
    OracleFallback,
    OrderCancellationData,
    OutcomePoint,
    OutcomePointEvent,
    OutcomePointTimeSeries,
    OutcomeSpace,
    OutcomeSpaceScalar,
    OutcomeSpaceTimeSeries,
    Timestamp,
)
from .types import IPLD_LINK, AliasedModel, Model


class PaginationFilter(Model):
    batch: Annotated[None | int, Field(gt=0, exclude=True)]
    batch_size: Annotated[None | int, Field(gt=0, exclude=True)]
    newest_first: Annotated[None | bool, Field(exclude=True)]

    @computed_field
    @property
    def page(self) -> None | int:
        return self.batch

    @computed_field
    @property
    def page_size(self) -> None | int:
        return self.batch_size

    @computed_field
    @property
    def sort(self) -> None | Literal["ASC", "DESC"]:
        match self.newest_first:
            case None:
                return None
            case True:
                return "DESC"
            case False:
                return "ASC"


# Authentication


class LoginSubmission(AliasedModel):
    message: str
    signature: str


class ExchangeParameters(AliasedModel):
    trading_protocol_id: str
    maker_trading_fee_rate: Decimal
    taker_trading_fee_rate: Decimal


# Admin API


class ExchangeProductListingSubmission(AliasedModel):
    id: Annotated[str, AfterValidator(validators.validate_hexstr32)]


class ExchangeProductUpdateSubmission(AliasedModel):
    listing_state: ListingState


# Trading API


class ExchangeProductFilter(PaginationFilter):
    pass


class OrderFilter(PaginationFilter):
    intent_account_id: str
    product_id: None | Annotated[str, AfterValidator(validators.validate_hexstr32)]
    type: None | OrderType
    states: Annotated[list[OrderState], Field(exclude=True)]
    side: None | OrderSide
    start: None | Timestamp
    end: None | Timestamp

    @computed_field
    @property
    def state(self) -> str | None:
        return ",".join(self.states) if self.states else None


class OrderSubmission(AliasedModel):
    type: OrderType
    intent: Intent | None = None
    cancellation_data: OrderCancellationData | None = None


class OrderFillFilter(PaginationFilter):
    intent_account_id: str
    product_id: None | Annotated[str, AfterValidator(validators.validate_hexstr32)]
    intent_hash: None | Annotated[str, AfterValidator(validators.validate_hexstr32)]
    start: None | Timestamp
    end: None | Timestamp
    trade_states: Annotated[list[TradeState], Field(exclude=True)]

    @computed_field
    @property
    def trade_state(self) -> str | None:
        return ",".join(self.trade_states) if self.trade_states else None


# IPFS Client


class ExtendedMetadata(Model):
    outcome_space: OutcomeSpaceTimeSeries | OutcomeSpaceScalar | OutcomeSpace
    outcome_point: OutcomePointEvent | OutcomePointTimeSeries | OutcomePoint
    oracle_config: OracleConfigPrototype1 | OracleConfig
    oracle_fallback: OracleFallback


class ComponentLink(AliasedModel):
    model_config = AliasedModel.model_config | ConfigDict(arbitrary_types_allowed=True)

    data: IPLD_LINK
    schema_: Annotated[IPLD_LINK, Field(alias="schema")]


class ExtendedMetadataDAG(Model):
    outcome_space: ComponentLink
    outcome_point: ComponentLink
    oracle_config: ComponentLink
    oracle_fallback: ComponentLink



"""MarginAccountRegistry contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract


class MarginAccountRegistry:
    """MarginAccountRegistry contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarginAccountRegistry contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the MarginAccountRegistry contract."""
        return self._contract.events.Initialized

    @property
    def MarginAccountCreated(self) -> contract.ContractEvent:
        """Binding for `event MarginAccountCreated` on the MarginAccountRegistry contract."""
        return self._contract.events.MarginAccountCreated

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the MarginAccountRegistry contract."""
        return self._contract.events.OwnershipTransferred

    @property
    def Upgraded(self) -> contract.ContractEvent:
        """Binding for `event Upgraded` on the MarginAccountRegistry contract."""
        return self._contract.events.Upgraded

    def upgrade_interface_version(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> str:
        """Binding for `UPGRADE_INTERFACE_VERSION` on the MarginAccountRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.UPGRADE_INTERFACE_VERSION().call(
            block_identifier=block_identifier
        )
        return str(return_value)

    def beacon(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `beacon` on the MarginAccountRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.beacon().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def clearing(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the MarginAccountRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def get_margin_account(
        self,
        collateral_asset: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getMarginAccount` on the MarginAccountRegistry contract.

        Parameters
        ----------
        collateral_asset : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getMarginAccount(
            collateral_asset,
        ).call(block_identifier=block_identifier)
        return eth_typing.ChecksumAddress(return_value)

    def initialize(
        self,
        clearing_: eth_typing.ChecksumAddress,
        beacon_: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the MarginAccountRegistry contract.

        Parameters
        ----------
        clearing_ : eth_typing.ChecksumAddress
        beacon_ : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            clearing_,
            beacon_,
        )

    def initialize_margin_account(
        self,
        collateral_asset: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initializeMarginAccount` on the MarginAccountRegistry contract.

        Parameters
        ----------
        collateral_asset : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initializeMarginAccount(
            collateral_asset,
        )

    def owner(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the MarginAccountRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def proxiable_uuid(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> hexbytes.HexBytes:
        """Binding for `proxiableUUID` on the MarginAccountRegistry contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.proxiableUUID().call(
            block_identifier=block_identifier
        )
        return hexbytes.HexBytes(return_value)

    def renounce_ownership(
        self,
    ) -> contract.ContractFunction:
        """Binding for `renounceOwnership` on the MarginAccountRegistry contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceOwnership()

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the MarginAccountRegistry contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )

    def upgrade_to_and_call(
        self,
        new_implementation: eth_typing.ChecksumAddress,
        data: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `upgradeToAndCall` on the MarginAccountRegistry contract.

        Parameters
        ----------
        new_implementation : eth_typing.ChecksumAddress
        data : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.upgradeToAndCall(
            new_implementation,
            data,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "UPGRADE_INTERFACE_VERSION",
            "inputs": [],
            "outputs": [{"name": "", "type": "string", "internalType": "string"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "beacon",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "clearing",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getMarginAccount",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IMarginAccount",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "initialize",
            "inputs": [
                {"name": "clearing_", "type": "address", "internalType": "address"},
                {"name": "beacon_", "type": "address", "internalType": "address"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "initializeMarginAccount",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IMarginAccount",
                }
            ],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "proxiableUUID",
            "inputs": [],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "renounceOwnership",
            "inputs": [],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "transferOwnership",
            "inputs": [
                {"name": "newOwner", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "upgradeToAndCall",
            "inputs": [
                {
                    "name": "newImplementation",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [],
            "stateMutability": "payable",
        },
        {
            "type": "event",
            "name": "Initialized",
            "inputs": [
                {
                    "name": "version",
                    "type": "uint64",
                    "indexed": False,
                    "internalType": "uint64",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "MarginAccountCreated",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "marginAccount",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "OwnershipTransferred",
            "inputs": [
                {
                    "name": "previousOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "newOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "Upgraded",
            "inputs": [
                {
                    "name": "implementation",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "AddressEmptyCode",
            "inputs": [
                {"name": "target", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "AlreadyInitialized", "inputs": []},
        {
            "type": "error",
            "name": "ERC1967InvalidImplementation",
            "inputs": [
                {"name": "implementation", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "ERC1967NonPayable", "inputs": []},
        {"type": "error", "name": "FailedCall", "inputs": []},
        {"type": "error", "name": "InvalidInitialization", "inputs": []},
        {"type": "error", "name": "NotInitializing", "inputs": []},
        {
            "type": "error",
            "name": "OwnableInvalidOwner",
            "inputs": [{"name": "owner", "type": "address", "internalType": "address"}],
        },
        {
            "type": "error",
            "name": "OwnableUnauthorizedAccount",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "UUPSUnauthorizedCallContext", "inputs": []},
        {
            "type": "error",
            "name": "UUPSUnsupportedProxiableUUID",
            "inputs": [{"name": "slot", "type": "bytes32", "internalType": "bytes32"}],
        },
    ],
)
